#
#    Copyright 2019 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

from airflow.plugins_manager import AirflowPlugin

from odahuflow.airflow_plugin.connection import GcpConnectionToOdahuConnectionOperator
from odahuflow.airflow_plugin.deployment import DeploymentOperator, DeploymentSensor
from odahuflow.airflow_plugin.api import OdahuHook
from odahuflow.airflow_plugin.model import ModelPredictRequestOperator, ModelInfoRequestOperator
from odahuflow.airflow_plugin.packaging import PackagingOperator, PackagingSensor
from odahuflow.airflow_plugin.training import TrainingOperator, TrainingSensor


class OdahuPlugin(AirflowPlugin):
    name = 'odahuflow'
    operators = [TrainingOperator, DeploymentOperator, PackagingOperator, ModelPredictRequestOperator,
                 ModelInfoRequestOperator, GcpConnectionToOdahuConnectionOperator]
    hooks = [OdahuHook]
    sensors = [TrainingSensor, DeploymentSensor, PackagingSensor]



#!/usr/bin/env python
#
#   Copyright 2017 EPAM Systems
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
"""
Tool for updating version files
"""
import argparse
import json
import re
import time
from typing import Sequence

# Constants
import semver
import yaml

PROFILE_VERSION_KEYS = [
    "odahu_infra_version",
    "odahuflow_version",
    "mlflow_toolchain_version",
    "jupyterlab_version",
    "packager_version",
    "examples_version",
    "odahu_airflow_plugin_version",
    "odahu_automation_version",
    "odahu_ui_version",
]

PY_SEARCH_PATTERN = r'''^__version__\s+=\s+'([0-9\.\-]+\w*?)'$'''
ODAHUFLOW_CONNECTIONS_KEY = "odahuflow_connections"
ODAHUFLOW_EXAMPLE_CONNECTION_ID = "odahu-flow-examples"


# Version files API


class VersionFile:
    """
    Interface that should be implemented by VersionFile Implementations
    """

    def __init__(self, fp):
        if not self.check_valid_file(fp):
            raise RuntimeError('Initialization of version file with not a valid format')
        self.fp = fp

    @staticmethod
    def check_valid_file(fp: str) -> bool:
        """
        Check whether this file is version file that is handled by this class or not
        :param fp: filepath to version file
        :return: True – means that this file can be handled by class, False – otherwise
        """
        raise NotImplementedError

    def set_version(self, version: str, **extra) -> None:
        """
        Version
        :param version: version
        :param extra: insert meta data to version file (optionally supported by different implementations)
        could be ignored
        :return: None
        """
        raise NotImplementedError

    def get_version(self):
        """
        Extract version from file
        :return:
        """
        raise NotImplementedError


class PythonVersionFile(VersionFile):
    """
    Handler for python `version.py`
    """

    @staticmethod
    def check_valid_file(fp: str) -> bool:
        return fp.endswith('version.py') or fp.endswith('version.info') or fp.endswith('__version__.py')

    def set_version(self, version, **extra):
        with open(self.fp, 'r') as stream:
            content = stream.read()
            if isinstance(content, bytes):
                content = content.decode('utf-8')

        # Set new __version__
        content = re.sub(
            PY_SEARCH_PATTERN, '__version__ = \'{}\''.format(version), content, flags=re.MULTILINE
        )

        with open(self.fp, 'w') as stream:
            stream.write(content)

    def get_version(self) -> str:
        with open(self.fp, 'r') as stream:
            try:
                content = stream.read()
                base_version = re.search(PY_SEARCH_PATTERN, content, flags=re.MULTILINE)
            except Exception:
                raise Exception('Can\'t get version from version string')

        return base_version.group(1)


class ProfileVersionFile(VersionFile):
    """
    Handler for internal profiles repo
    """

    @staticmethod
    def check_valid_file(fp: str) -> bool:
        return fp.endswith('common.yaml')

    def set_version(self, version, **extra):
        with open(self.fp, 'r') as f:
            data = yaml.load(f)

        for key in PROFILE_VERSION_KEYS:
            data[key] = version

        # Profiles have to contain an Odahu-flow connection that points to the examples repository.
        # For every release, we create a tag with version name for the examples repository.
        # In addition to the main versions, we also change the reference of the connection.
        odahuflow_connections = data[ODAHUFLOW_CONNECTIONS_KEY]
        for conn in odahuflow_connections:
            if conn['id'] == ODAHUFLOW_EXAMPLE_CONNECTION_ID:
                conn['spec']['reference'] = version

        with open(self.fp, 'w') as f:
            yaml.dump(data, f)

    def get_version(self) -> str:
        with open(self.fp, 'r') as f:
            data = yaml.load(f)

        return data[PROFILE_VERSION_KEYS[0]]


class NpmVersionFile(VersionFile):
    """
    Handler for npm `package.json`
    """

    @staticmethod
    def check_valid_file(fp: str) -> bool:
        return fp.endswith('package.json')

    def set_version(self, version: str, **extra):
        with open(self.fp) as f:
            package = json.load(f)

        package['version'] = version

        with open(self.fp, 'w') as f:
            json.dump(package, f, indent=2, sort_keys=True)

    def get_version(self):
        with open(self.fp) as f:
            package = json.load(f)

        return package['version']


ver_file_classes = [PythonVersionFile, NpmVersionFile, ProfileVersionFile]


def look_version_file_cls(fp: str):
    """
    Looks `VersionClass` that can handle file at filepath `fp`
    :param fp: filepath
    :return: `VersionClass` implementation if found – None otherwise
    """
    for cls in ver_file_classes:
        if cls.check_valid_file(fp):
            return cls
    else:
        raise ValueError(f'Unknown type of the file: {fp}')


def init_ver_file(fp: str) -> VersionFile:
    """
    Looks appropriate `VersionFile` class for file at given filepath (`fp`)
    Raise `RuntimeException` if class was not found
    :param fp:
    :return:
    """

    cls = look_version_file_cls(fp)
    if not cls:
        raise RuntimeError(f'VersionFile class for file {fp} is not found')

    return cls(fp)


def get_version_files(fps: Sequence[str]) -> Sequence[VersionFile]:
    """
    Initialize `VersionFile` instances by file paths
    :param fps: file paths to version files
    :return: Sequence[VersionFile]
    """

    ver_files = []

    for fp in fps:
        ver_files.append(init_ver_file(fp))

    return ver_files


# Script Arguments


class ScriptArgs:

    def __init__(
            self, build_id: int, build_user: str, date_string: str,
            git_revision: str, build_version: str, fps: Sequence[str]
    ):
        """
        :param build_id: Build serial number that is assigned by Jenkins for each job build
        :param build_user: User who launch job build
        :param date_string: Date when build was launched
        :param git_revision: Git revision
        :param build_version: Explicitly defined build version
        :param fps: version file paths
        """
        self.build_id = build_id
        self.build_user = build_user
        self.date_string = date_string
        self.git_revision = git_revision
        self.build_version = build_version
        self.fps = fps


def set_defaults(script_args: ScriptArgs, ver_file_for_meta: VersionFile):
    """
    Set defaults to script args
    :param script_args: Script args
    :param ver_file_for_meta: version file from which some meta data about current version will be extracted
    :return: None
    """
    if not script_args.build_version:
        local_version_string = f'b{int(round(time.time() * 1000))}'
        finalized_version = semver.finalize_version(ver_file_for_meta.get_version())

        script_args.build_version = '{}-{}'.format(finalized_version, local_version_string)


# MAIN

def main(args) -> str:
    version_files = get_version_files(args.fps)

    if not version_files:
        print('At least one version file must be specified')
        exit(3)

    script_args = ScriptArgs(
        args.build_id, args.build_user, args.date_string,
        args.git_revision, args.build_version, args.fps
    )
    set_defaults(script_args, ver_file_for_meta=version_files[0])

    try:
        for version_file in version_files:
            version_file.set_version(
                version=script_args.build_version,
                build_id=script_args.build_id,
                git_revision=script_args.git_revision,
                build_user=script_args.build_user
            )

    except KeyboardInterrupt:
        print('Interrupt')
        exit(2)

    return script_args.build_version


if __name__ == '__main__':
    parser = argparse.ArgumentParser('Version file updater (adds time, build id, commit id to version)')
    parser.add_argument('build_id', type=int, help='Set build id')
    parser.add_argument('build_user', type=str, help='Set build user')
    parser.add_argument('date_string', type=str, help='Set date string')
    parser.add_argument('git_revision', type=str, help='Set git revision')
    parser.add_argument('fps', nargs='+', help='Paths to version files')
    parser.add_argument('--build-version', type=str, help='Explicitly specify new Legion build version')

    args = parser.parse_args()

    res = main(args)
    print(res)



"""
This module provide API to build odahu models (ModelTraining, ModelConnection, etc.) from yaml/json files
"""
import logging
import os
from io import StringIO
from typing import Union, Tuple

from airflow import settings
from odahuflow.sdk.clients.api_aggregated import OdahuflowCloudResourceUpdatePair, parse_stream

logger = logging.getLogger(__name__)


def _find_absolute_resource_file(path: str) -> Union[str, None]:
    """
    Try to find absolute path for resource file
    :param path: str relative path to (the most foreground is top):
        * Airflow DAGs
        * Airflow HOME
    :return: absolute path as :str if file exists otherwise `None`
    """

    base_paths = [settings.DAGS_FOLDER, settings.AIRFLOW_HOME]

    for bp in base_paths:
        abs_path = os.path.join(bp, path)
        if os.path.exists(abs_path) and os.path.isfile(abs_path):
            return abs_path


def _build_resource_body(path_or_body: str, **template_vars) -> Tuple[str, Union[str, None]]:
    """
    1. If path_or_body is a valid filepath to odahu resource YAML/JSON file then it loads its content
    2. If path_or_body is a YAML/JSON string that describe odahu resource then it handles it as is
    3. Render content as `path_or_body.format(**template_vars)`

    :param path_or_body: could be
        * a relative filepath to DAGs or HOME Airflow folders to odahu resource YAML/JSON file
        * a YAML/JSON string that describe Odahu resource
    :param template_vars: variables that could be used to render resource body (if it contains bracket rounded pieces)
    :return: Odahu resource body as a YAML/JSON string and absolute filepath to definition file
    if this body was read from file (otherwise None)
    """

    abs_path_to_resource_file = _find_absolute_resource_file(path_or_body)

    if abs_path_to_resource_file:
        with open(abs_path_to_resource_file, 'r') as data_stream:
            resource_body_tpl = data_stream.read()
    else:
        resource_body_tpl = path_or_body

    try:
        resource_body = resource_body_tpl.format(**template_vars)
    except KeyError as exc_info:
        raise KeyError('Not all template variables in payload have values. '
                       'Please pass values for all vars as keyword arguments') from exc_info

    return resource_body, abs_path_to_resource_file


def _parse_resource_body(resource_body: str) -> OdahuflowCloudResourceUpdatePair:
    """
    Build `OdahuflowCloudResourceUpdatePair` from YAML/JSON content string
    :param resource_body:  YAML/JSON odahu resource definition
    :return: OdahuflowCloudResourceUpdatePair
    """

    resources = parse_stream(StringIO(resource_body))

    resources_len = len(resources.changes)
    if resources_len != 1:
        raise TypeError(f'Resource definition contains {resources_len} entities but must contains exactly one')

    resource_pair: OdahuflowCloudResourceUpdatePair = resources.changes[0]

    return resource_pair


def resource(path_or_body: str, **template_vars) -> OdahuflowCloudResourceUpdatePair:
    """
    Create `OdahuflowCloudResourceUpdatePair` object from resource definition in YAML/JSON format.
    Resource definition could be passed as a string in `path_or_body` OR a filepath to a valid YAML/JSON file.
    :param path_or_body: could be
        * a relative filepath to DAGs or HOME Airflow folders to odahu resource YAML/JSON file
        * a YAML/JSON string that describe Odahu resource
    :param template_vars: variables that could be used to render resource body (if it contains bracket rounded pieces)
    :return: OdahuflowCloudResourceUpdatePair with resource that was build from resource definition
    """

    resource_body, fp = _build_resource_body(path_or_body, **template_vars)

    pair = _parse_resource_body(resource_body)

    logger.info(f'Resource {pair.resource} was initialized from {"string" if fp is None else fp}')

    return pair



#
#    Copyright 2019 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

import typing

from airflow.models import BaseOperator
from airflow.utils.decorators import apply_defaults
from odahuflow.sdk.clients.model import ModelClient

from odahuflow.airflow_plugin.api import OdahuHook


class ModelPredictRequestOperator(BaseOperator):

    @apply_defaults
    def __init__(self,
                 model_deployment_name: str,
                 api_connection_id: str,
                 model_connection_id: str,
                 request_body: typing.Any,
                 md_role_name: str = "",
                 *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.model_deployment_name = model_deployment_name
        self.request_body = request_body
        self.model_connection_id = model_connection_id
        self.api_connection_id = api_connection_id
        self.md_role_name = md_role_name

    def get_hook(self) -> OdahuHook:
        return OdahuHook(
            self.api_connection_id,
            self.model_connection_id
        )

    def execute(self, context):
        # pylint: disable=unused-argument
        model_client: ModelClient = self.get_hook().get_model_client(self.model_deployment_name)

        resp = model_client.invoke(**self.request_body)
        print(resp)

        return resp


class ModelInfoRequestOperator(BaseOperator):

    @apply_defaults
    def __init__(self,
                 model_deployment_name: str,
                 api_connection_id: str,
                 model_connection_id: str,
                 md_role_name: str = "",
                 *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.model_deployment_name = model_deployment_name
        self.model_connection_id = model_connection_id
        self.api_connection_id = api_connection_id
        self.md_role_name = md_role_name

    def get_hook(self) -> OdahuHook:
        return OdahuHook(
            self.api_connection_id,
            self.model_connection_id
        )

    def execute(self, context):
        # pylint: disable=unused-argument
        model_client: ModelClient = self.get_hook().get_model_client(self.model_deployment_name)

        resp = model_client.info()
        print(resp)

        return resp



from datetime import datetime

from airflow import DAG
from airflow import settings
from odahuflow.airflow_plugin.deployment import DeploymentOperator, DeploymentSensor
from odahuflow.airflow_plugin.model import ModelPredictRequestOperator, ModelInfoRequestOperator
from odahuflow.airflow_plugin.packaging import PackagingOperator, PackagingSensor
from odahuflow.airflow_plugin.resources import resource
from odahuflow.airflow_plugin.training import TrainingOperator, TrainingSensor
from odahuflow.sdk.models import ModelDeployment, ModelDeploymentSpec

RES_DIR = 'tests/resources'


def test_hello_world():
    default_args = {
        'owner': 'airflow',
        'depends_on_past': False,
        'start_date': datetime(2019, 9, 3),
        'email_on_failure': False,
        'email_on_retry': False,
        'end_date': datetime(2099, 12, 31)
    }

    odahu_connection_api = "odahu_api"
    model_connection_id = "odahu_model"

    OLD_DAG_FOLDER = settings.DAGS_FOLDER
    settings.DAGS_FOLDER = RES_DIR
    try:
        training_id, training = resource('training.odahuflow.yaml')
    finally:
        settings.DAGS_FOLDER = OLD_DAG_FOLDER

    packaging_id, packaging = resource("""
    id: airlfow-wine
    kind: ModelPackaging
    spec:
      artifactName: "<fill-in>"
      targets:
        - connectionName: docker-ci
          name: docker-push
      integrationName: docker-rest
    """)

    deployment_id = "airlfow-tensorflow"
    deployment = ModelDeployment(
        id=deployment_id,
        spec=ModelDeploymentSpec(
            min_replicas=1,
        ),
    )

    model_example_request = {
        "columns": ["features"] * 13,
        "data": [
            [18.08460, 0.0, 18.10, 0.0, 0.6790, 6.434, 100.0, 1.8347, 24.0, 666.0, 20.2, 27.25, 29.05],
            [2.92400, 0.0, 19.58, 0.0, 0.6050, 6.101, 93.0, 2.2834, 5.0, 403.0, 14.7, 240.16, 9.81]
        ]
    }

    dag = DAG(
        'tensorflow_example_model',
        default_args=default_args,
        schedule_interval=None
    )

    with dag:
        train = TrainingOperator(
            task_id="training",
            api_connection_id=odahu_connection_api,
            training=training,
            default_args=default_args
        )

        wait_for_train = TrainingSensor(
            task_id='wait_for_training',
            training_id=training_id,
            api_connection_id=odahu_connection_api,
            default_args=default_args
        )

        pack = PackagingOperator(
            task_id="packaging",
            api_connection_id=odahu_connection_api,
            packaging=packaging,
            trained_task_id="wait_for_training",
            default_args=default_args
        )

        wait_for_pack = PackagingSensor(
            task_id='wait_for_packaging',
            packaging_id=packaging_id,
            api_connection_id=odahu_connection_api,
            default_args=default_args
        )

        dep = DeploymentOperator(
            task_id="deployment",
            api_connection_id=odahu_connection_api,
            deployment=deployment,
            packaging_task_id="wait_for_packaging",
            default_args=default_args
        )

        wait_for_dep = DeploymentSensor(
            task_id='wait_for_deployment',
            deployment_id=deployment_id,
            api_connection_id=odahu_connection_api,
            default_args=default_args
        )

        model_predict_request = ModelPredictRequestOperator(
            task_id="model_predict_request",
            model_deployment_name=deployment_id,
            api_connection_id=odahu_connection_api,
            model_connection_id=model_connection_id,
            request_body=model_example_request,
            default_args=default_args
        )

        model_info_request = ModelInfoRequestOperator(
            task_id='model_info_request',
            model_deployment_name=deployment_id,
            api_connection_id=odahu_connection_api,
            model_connection_id=model_connection_id,
            default_args=default_args
        )

        # pylint: disable=pointless-statement
        train >> wait_for_train >> pack >> wait_for_pack >> dep >> wait_for_dep
        wait_for_dep >> model_info_request
        wait_for_dep >> model_predict_request



#
#    Copyright 2019 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
import time

from airflow.models import BaseOperator
from airflow.operators.sensors import BaseSensorOperator
from airflow.utils.decorators import apply_defaults
from odahuflow.airflow_plugin.api import OdahuHook
from odahuflow.airflow_plugin.packaging import XCOM_PACKAGING_RESULT_KEY
from odahuflow.sdk.clients.api import WrongHttpStatusCode
from odahuflow.sdk.clients.deployment import ModelDeploymentClient, READY_STATE, FAILED_STATE
from odahuflow.sdk.models import ModelDeployment

DEFAULT_WAIT_TIMEOUT = 5


class DeploymentOperator(BaseOperator):

    @apply_defaults
    def __init__(self,
                 deployment: ModelDeployment,
                 api_connection_id: str,
                 packaging_task_id: str = "",
                 *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.deployment = deployment
        self.api_connection_id = api_connection_id
        self.packaging_task_id = packaging_task_id

    def get_hook(self) -> OdahuHook:
        return OdahuHook(
            self.api_connection_id
        )

    def execute(self, context):
        client: ModelDeploymentClient = self.get_hook().get_api_client(ModelDeploymentClient)

        try:
            if self.packaging_task_id:
                result = context['task_instance'].xcom_pull(task_ids=self.packaging_task_id,
                                                            key=XCOM_PACKAGING_RESULT_KEY)
                print(result)
                self.deployment.spec.image = result["image"]

            if self.deployment.id:
                client.delete(self.deployment.id)

                self.wait_delete_operation_finish(200, self.deployment.id, client)
        except WrongHttpStatusCode as e:
            if e.status_code != 404:
                raise e

        dep = client.create(self.deployment)

        return dep.id

    def wait_delete_operation_finish(self, timeout: int, md_id: str, md_client: ModelDeploymentClient):
        start = time.time()

        while True:
            elapsed = time.time() - start
            if elapsed > timeout:
                raise Exception('Time out: operation has not been confirmed')

            try:
                md_client.get(md_id)
            except WrongHttpStatusCode as e:
                if e.status_code == 404:
                    self.log.info(f'Model deployment {md_id} was deleted')

                    return
                self.log.info('Callback have not confirmed completion of the operation')

            self.log.info(f'Model deployment {md_id} is still being deleted...')
            time.sleep(DEFAULT_WAIT_TIMEOUT)


class DeploymentSensor(BaseSensorOperator):

    @apply_defaults
    def __init__(self,
                 deployment_id: str,
                 api_connection_id: str,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.deployment_id = deployment_id
        self.api_connection_id = api_connection_id

    def get_hook(self) -> OdahuHook:
        return OdahuHook(
            self.api_connection_id
        )

    def poke(self, context):
        # pylint: disable=unused-argument
        client: ModelDeploymentClient = self.get_hook().get_api_client(ModelDeploymentClient)

        dep_status = client.get(self.deployment_id).status

        if dep_status.state == FAILED_STATE:
            raise Exception(f'Deployment {self.deployment_id} was failed')

        return dep_status.state == READY_STATE



#
#    Copyright 2019 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

import json

import requests
from airflow.hooks.base_hook import BaseHook
from airflow.models import Connection
from odahuflow.sdk.clients.model import ModelClient
from odahuflow.sdk.clients.route import ModelRouteClient


class OdahuHook(BaseHook):  # pylint: disable=abstract-method

    def __init__(self, api_connection_id=None, model_connection_id=None):
        super().__init__(None)

        self.api_connection_id = api_connection_id
        self.model_connection_id = model_connection_id

    def get_api_client(self, target_client_class):
        api_conn = self.get_connection(self.api_connection_id)
        self.log.info(api_conn)

        return target_client_class(f'{api_conn.schema}://{api_conn.host}', self._get_token(api_conn))

    def get_model_client(self, model_route_name: str) -> ModelClient:
        api_conn = self.get_connection(self.api_connection_id)
        self.log.info(api_conn)

        model_conn = self.get_connection(self.model_connection_id)
        self.log.info(model_conn)

        mr_client: ModelRouteClient = self.get_api_client(ModelRouteClient)
        model_route = mr_client.get(model_route_name)

        return ModelClient(model_route.status.edge_url, self._get_token(api_conn))

    def _get_token(self, conn: Connection) -> str:
        """
        Authorize test user and get access token.

        :param Airlfow API connection
        :return: access token
        """
        extra = json.loads(conn.extra)

        auth_url = extra["auth_url"]

        try:
            response = requests.post(
                auth_url,
                data={
                    'grant_type': 'client_credentials',
                    'client_id': extra["client_id"],
                    'client_secret': extra["client_secret"],
                    'scope': extra['scope']
                }
            )
            response_data = response.json()

            # Parse fields and return
            id_token = response_data.get('id_token')
            token_type = response_data.get('token_type')
            expires_in = response_data.get('expires_in')

            self.log.info('Received %s token with expiration in %d seconds', token_type, expires_in)

            return id_token
        except requests.HTTPError as http_error:
            raise Exception(f'Can not authorize user {conn.login} on {auth_url}: {http_error}') from http_error



#
#    Copyright 2019 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

from airflow.models import BaseOperator
from airflow.operators.sensors import BaseSensorOperator
from airflow.utils.decorators import apply_defaults
from odahuflow.airflow_plugin.api import OdahuHook
from odahuflow.sdk.clients.api import WrongHttpStatusCode
from odahuflow.sdk.clients.training import ModelTrainingClient, TRAINING_SUCCESS_STATE, TRAINING_FAILED_STATE
from odahuflow.sdk.models import ModelTraining

XCOM_TRAINED_ARTIFACT_KEY = "trained_artifact_name"


class TrainingOperator(BaseOperator):

    @apply_defaults
    def __init__(self,
                 training: ModelTraining,
                 api_connection_id: str,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.training = training
        self.api_connection_id = api_connection_id

    def get_hook(self) -> OdahuHook:
        return OdahuHook(
            self.api_connection_id
        )

    def execute(self, context):
        # pylint: disable=unused-argument
        client: ModelTrainingClient = self.get_hook().get_api_client(ModelTrainingClient)

        try:
            if self.training.id:
                client.delete(self.training.id)
        except WrongHttpStatusCode as e:
            if e.status_code != 404:
                raise e

        train = client.create(self.training)

        return train.id


class TrainingSensor(BaseSensorOperator):

    @apply_defaults
    def __init__(self,
                 training_id: str,
                 api_connection_id: str,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.training_id = training_id
        self.api_connection_id = api_connection_id

    def get_hook(self) -> OdahuHook:
        return OdahuHook(
            self.api_connection_id
        )

    def poke(self, context):
        client: ModelTrainingClient = self.get_hook().get_api_client(ModelTrainingClient)

        train_status = client.get(self.training_id).status

        if train_status.state == TRAINING_FAILED_STATE:
            raise Exception(f'Model training {self.training_id} was failed')

        if train_status.state == TRAINING_SUCCESS_STATE:
            context['task_instance'].xcom_push(key=XCOM_TRAINED_ARTIFACT_KEY,
                                               value=train_status.artifacts[-1].artifact_name)

            return True

        return False



#
#    Copyright 2019 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

from airflow.models import BaseOperator
from airflow.operators.sensors import BaseSensorOperator
from airflow.utils.decorators import apply_defaults
from odahuflow.airflow_plugin.api import OdahuHook
from odahuflow.airflow_plugin.training import XCOM_TRAINED_ARTIFACT_KEY
from odahuflow.sdk.clients.api import WrongHttpStatusCode
from odahuflow.sdk.clients.packaging import ModelPackagingClient, SUCCEEDED_STATE, FAILED_STATE
from odahuflow.sdk.models import ModelPackaging

XCOM_PACKAGING_RESULT_KEY = "packaging_result"


class PackagingOperator(BaseOperator):

    @apply_defaults
    def __init__(self,
                 packaging: ModelPackaging,
                 api_connection_id: str,
                 trained_task_id: str = "",
                 *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.packaging = packaging
        self.api_connection_id = api_connection_id
        self.trained_task_id = trained_task_id

    def get_hook(self) -> OdahuHook:
        return OdahuHook(
            self.api_connection_id
        )

    def execute(self, context):
        client: ModelPackagingClient = self.get_hook().get_api_client(ModelPackagingClient)

        try:
            if self.trained_task_id:
                artifact_name = context['task_instance'].xcom_pull(task_ids=self.trained_task_id,
                                                                   key=XCOM_TRAINED_ARTIFACT_KEY)
                print(artifact_name)
                self.packaging.spec.artifact_name = artifact_name

            if self.packaging.id:
                client.delete(self.packaging.id)
        except WrongHttpStatusCode as e:
            if e.status_code != 404:
                raise e

        train = client.create(self.packaging)

        return train.id


class PackagingSensor(BaseSensorOperator):

    @apply_defaults
    def __init__(self,
                 packaging_id: str,
                 api_connection_id: str,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.packaging_id = packaging_id
        self.api_connection_id = api_connection_id

    def get_hook(self) -> OdahuHook:
        return OdahuHook(
            self.api_connection_id
        )

    def poke(self, context):
        client: ModelPackagingClient = self.get_hook().get_api_client(ModelPackagingClient)

        pack_status = client.get(self.packaging_id).status

        if pack_status.state == FAILED_STATE:
            raise Exception(f'Model packaging {self.packaging_id} was failed')

        if pack_status.state == SUCCEEDED_STATE:
            results = {result.name: result.value for result in pack_status.results}

            context['task_instance'].xcom_push(XCOM_PACKAGING_RESULT_KEY, results)

            return True

        return False



import json
import os
import re
from collections import namedtuple

from tools import update_version_files

BASE_VER = '1.0.0'
PY_SEARCH_PATTERN = r'''^__version__\s+=\s+'([0-9\.\-]+\w*?)'$'''

args_cls = namedtuple('args', 'build_id, build_user, date_string, build_version, git_revision, use_full_commit_id, fps')


def get_py_paths(main_path):

    return [
        f'{main_path}/legion/cli/legion/cli/',
        f'{main_path}/legion/sdk/legion/sdk/',
        f'{main_path}/legion/jupyterlab-plugin/legion/jupyterlab/',
    ]


def get_npm_paths(main_path):

    return [
        f'{main_path}/legion/jupyterlab-plugin/'
    ]


def get_info_paths(main_path):

    return [
        f'{main_path}/'
    ]


def get_fps(main_path):

    fps = []

    for py_fp in get_py_paths(main_path):
        fp = f'{py_fp}/version.py'
        fps.append(fp)

    for npm_fp in get_npm_paths(main_path):
        fp = f'{npm_fp}/package.json'
        fps.append(fp)

    for npm_fp in get_info_paths(main_path):
        fp = f'{npm_fp}/version.info'
        fps.append(fp)

    return fps


def create_structure(main_path):

    py_paths = get_py_paths(main_path)
    npm_paths = get_npm_paths(main_path)
    info_paths = get_info_paths(main_path)

    for path in py_paths:
        os.makedirs(path, exist_ok=True)
        with open(f'{path}/version.py', 'w') as f:
            f.write(f"__version__ = '{BASE_VER}'\n")

    for path in npm_paths:
        os.makedirs(path, exist_ok=True)
        package = {'version': f'{BASE_VER}'}
        with open(f'{path}/package.json', 'w') as f:
            json.dump(package, f)

    for path in info_paths:
        os.makedirs(path, exist_ok=True)
        with open(f'{path}/version.info', 'w') as f:
            f.write(f"__version__ = '{BASE_VER}'\n")


def assert_expected_version(version: str, expected_version: str, main_path: str):
    """
    Check that version was properly calculated and version files were updated
    :param version: version that is calculated by script
    :param expected_version: version that is expected
    :param main_path: project directory where version files should be updated
    :return:
    """

    assert version == expected_version

    for dir_ in get_npm_paths(main_path):
        fp = f'{dir_}/package.json'
        with open(fp) as f:
            package = json.load(f)
            assert package['version'] == expected_version

    for dir_ in get_py_paths(main_path):
        fp = f'{dir_}/version.py'
        with open(fp) as f:
            content = f.read()
            base_version = re.search(PY_SEARCH_PATTERN, content, flags=re.MULTILINE)
            ver = base_version.group(1)
            assert ver == expected_version

    for dir_ in get_info_paths(main_path):
        fp = f'{dir_}/version.info'
        with open(fp) as f:
            content = f.read()
            base_version = re.search(PY_SEARCH_PATTERN, content, flags=re.MULTILINE)
            ver = base_version.group(1)
            assert ver == expected_version


def test_update_version_id_stable(tmp_path):

    # Given

    create_structure(tmp_path)

    # When
    os.chdir(tmp_path)
    fps = get_fps(tmp_path)
    args = args_cls(17, 'null', '20190513113315', '2.0.0-rc1', '0000', False, fps)
    expected_version = args.build_version

    version = update_version_files.main(args)

    # Then
    assert_expected_version(version, expected_version, tmp_path)


def test_update_version_id_not_stable(tmp_path):

    # Given

    create_structure(tmp_path)

    # When
    os.chdir(tmp_path)
    fps = get_fps(tmp_path)
    args = args_cls(17, 'null', '20190513113315', None, '0000', False, fps)
    expected_version = f'{BASE_VER}-dev{args.build_id}'

    version = update_version_files.main(args)

    # Then
    assert_expected_version(version, expected_version, tmp_path)




# -*- coding: utf-8 -*-
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from builtins import range
from datetime import timedelta

from airflow.models import DAG
from airflow.operators.bash_operator import BashOperator
from airflow.utils.dates import days_ago

args = {
    'owner': 'Airflow',
    'start_date': days_ago(1),
}

dag = DAG(
    dag_id='health_check',
    catchup=False,
    default_args=args,
    schedule_interval='*/5 * * * *',
    dagrun_timeout=timedelta(seconds=20),
)

run_this = BashOperator(
    task_id='run',
    bash_command="echo healthy!",
    dag=dag,
)

run_this

if __name__ == "__main__":
    dag.cli()




#!/usr/bin/env python3
#
#    Copyright 2019 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
import os
import re

from setuptools import setup, find_namespace_packages

PACKAGE_ROOT_PATH = os.path.dirname(os.path.abspath(__file__))
VERSION_FILE = os.path.join(PACKAGE_ROOT_PATH, 'odahuflow/airflow_plugin', '__version__.py')


def extract_version() -> str:
    """
    Extract version from .py file using regex

    :return: odahuflow version
    """
    with open(VERSION_FILE, 'rt') as version_file:
        file_content = version_file.read()
        VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
        mo = re.search(VSRE, file_content, re.M)
        if mo:
            return mo.group(1)
        else:
            raise RuntimeError("Unable to find version string in %s." % (file_content,))


setup(
    name='odahu-flow-airflow-plugin',
    author='Vlad Tokarev, Vitalik Solodilov',
    author_email='vlad.tokarev.94@gmail.com, mcdkr@yandex.ru',
    license='Apache v2',
    classifiers=[
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    keywords='airflow odahuflow',
    python_requires='>=3.6',
    packages=find_namespace_packages(),
    data_files=[('', ["README.md"])],
    zip_safe=False,
    entry_points={
        'airflow.plugins': [
            'odahuflow = odahuflow.airflow_plugin.plugin:OdahuPlugin'
        ]
    },
    install_requires=[
        'odahu-flow-sdk==1.1.0rc9',
        'requests>=2.22.0',
        # because this package is in conflict of odahu-flow-sdk and requests but pip doesn't have true dependency
        # resolver (https://github.com/pypa/pip/issues/988)
        'urllib3>=1.24.3'
    ],
    extras_require={
        'testing': [
            'pytest>=5.1.2',
            'pytest-mock>=1.10.4',
            'pytest-cov>=2.7.1',
            'pylint>=2.3.0',
            'apache-airflow==1.10.2',
            'google-api-python-client'
        ]
    },
    version=extract_version(),
)



#
#    Copyright 2019 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

from airflow.models import BaseOperator, Connection as AirflowConnection
from airflow.utils.decorators import apply_defaults
from odahuflow.airflow_plugin.api import OdahuHook
from odahuflow.sdk.clients.api import WrongHttpStatusCode
from odahuflow.sdk.clients.connection import ConnectionClient
from odahuflow.sdk.models import Connection

GCP_CREDENTIALS_KEYFILE_DICT = "extra__google_cloud_platform__keyfile_dict"


class GcpConnectionToOdahuConnectionOperator(BaseOperator):

    @apply_defaults
    def __init__(self,
                 api_connection_id: str,
                 google_cloud_storage_conn_id: str,
                 conn_template: Connection,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.conn_template = conn_template
        self.api_connection_id = api_connection_id
        self.google_cloud_storage_conn_id = google_cloud_storage_conn_id

    def get_hook(self) -> OdahuHook:
        return OdahuHook(
            self.api_connection_id
        )

    def execute(self, context):
        gcp_conn: AirflowConnection = self.get_hook().get_connection(self.google_cloud_storage_conn_id)
        self.conn_template.spec.key_secret = gcp_conn.extra_dejson[GCP_CREDENTIALS_KEYFILE_DICT]
        client: ConnectionClient = self.get_hook().get_api_client(ConnectionClient)

        try:
            if self.conn_template.id:
                client.delete(self.conn_template.id)
        except WrongHttpStatusCode as e:
            if e.status_code != 404:
                raise e

        conn = client.create(self.conn_template)

        self.log.info(conn)
